import openai_responses
from pydantic import BaseModel
import pytest
import openai
from moorellm import MooreFSM, MooreState, DefaultResponse
from moorellm.models import MooreRun, ImmediateStateChange
from moorellm.utils import wrap_into_json_response


# Define the fixture
@pytest.fixture
def fsm():
    """Fixture for creating a MooreFSM instance."""
    return MooreFSM(initial_state="START")


@pytest.fixture
def openai_client():
    return openai.AsyncOpenAI(api_key="sk-fake123")


@pytest.fixture
def set_openai_response():
    def set_response(
        openai_mock: openai_responses.OpenAIMock, response: BaseModel, next_state: str
    ):
        openai_mock.beta.chat.completions.create.response = {
            "choices": [
                {
                    "index": 0,
                    "finish_reason": "stop",
                    "message": {
                        "content": wrap_into_json_response(response, next_state),
                        "role": "assistant",
                    },
                }
            ]
        }

    return set_response


# Test the module creation
def test_module_creation(fsm):
    """Test that the module can be created."""
    assert isinstance(fsm, MooreFSM)


# Test for simple fsm with state
@pytest.mark.asyncio
@openai_responses.mock()
async def test_immediete_state_transition(
    fsm: MooreFSM,
    openai_client: openai.AsyncOpenAI,
    set_openai_response,
    openai_mock: openai_responses.OpenAIMock,
):
    @fsm.state(
        state_key="START",
        system_prompt="After this Change state but not immediately",
        transitions={
            "CHANGE_NOT_IMMEDIATE": "change not immediately",
            "CHANGE_IMMEDIATE": "Changing state immediately"
        },
    )
    async def start_state(fsm: MooreFSM, response: str, will_transition: bool):
        if (will_transition and fsm.get_next_state() == "CHANGE_IMMEDIATE"):
            # Logic to do urgent state change with response from next state :)
            return ImmediateStateChange(next_state="CHANGE_IMMEDIATE")

        return "I REPLIED FROM START_STATE"
    
    @fsm.state(
        state_key="CHANGE_NOT_IMMEDIATE",
        system_prompt="Change state but not immediately",
        transitions={
            "START": "Change to start state"
        },
    )
    async def change_not_immediate_state(fsm: MooreFSM, response: str, will_transition: bool):
        return "I REPLIED FROM CHANGE_NOT_IMMEDIATE_STATE"
    
    @fsm.state(
        state_key="CHANGE_IMMEDIATE",
        system_prompt="Change state immediately",
        transitions={
            "START": "Change to start state"
        },
    )
    async def change_immediate_state(fsm: MooreFSM, response: str, will_transition: bool):
        return "I REPLIED FROM CHANGE_IMMEDIATE_STATE"

    set_openai_response(
        openai_mock, DefaultResponse(content=""), next_state="CHANGE_NOT_IMMEDIATE"
    )

    # Run the FSM
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "CHANGE_NOT_IMMEDIATE"
    # First the response should be from the start state (as we made simple transition)
    assert run_state.response == "I REPLIED FROM START_STATE"

    # For sake of test, set the next state to be main again
    fsm._state = "START"

    set_openai_response(
        openai_mock, DefaultResponse(content=""), next_state="CHANGE_IMMEDIATE"
    )

    # Now test the transitions in case immediate state change
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "CHANGE_IMMEDIATE"
    # Now it should reply from the immediate state
    assert run_state.response == "I REPLIED FROM CHANGE_IMMEDIATE_STATE"


from typing import Union
import openai
import jinja2
from pydantic import BaseModel
from moorellm.models import GuardrailResponse

# ------ Define Helper Functions ------ #
async def structured_call(chat_history: list, openai_instance: Union[openai.AzureOpenAI, openai.AsyncAzureOpenAI], response_model: type[BaseModel], model_name: str = "gpt-4o-2024-08-06", **kwargs):
    completion: openai.types.Completion = await openai_instance.beta.chat.completions.parse(
        model=model_name,
        messages=chat_history,
        response_format=response_model,
    )

    message = completion.choices[0].message
    if message.parsed:
        message: response_model = message.parsed
        return message
    else:
        print(message.refusal)
        raise Exception("Failed to get structured_call, please try again.." + message.refusal)

# Guardrail transition function for moorellm
async def guardrail_transition(condition: str, state_from: str, state_to: str, chat_history: list[dict], openai_instance: Union[openai.AzureOpenAI, openai.AsyncAzureOpenAI]):
    system_prompt = f"""You are an AI states management system responsible for checking if transitions
                        from one state to another are valid, You observed that a state transition was requested
                        from state {state_from} to state {state_to}, Use user conversation upto this point to verify
                        if this transition is valid or not.

                        The condition for this transition is: {condition}"""

    chat_history_copy = chat_history.copy()
    # Iterate through the chat history and remove the system prompts
    for chat in chat_history_copy:
        if chat["role"] == "system":
            chat_history.remove(chat)
    # Add the system prompt
    chat_history.append({"role": "system", "content": system_prompt})

    response: GuardrailResponse = await structured_call(chat_history, openai_instance, GuardrailResponse)
    return response


import openai_responses
from pydantic import BaseModel
import pytest
import openai
from moorellm import MooreFSM, MooreState, DefaultResponse
from moorellm.models import MooreRun
from moorellm.utils import wrap_into_json_response


# Define the fixture
@pytest.fixture
def fsm():
    """Fixture for creating a MooreFSM instance."""
    return MooreFSM(initial_state="START")


@pytest.fixture
def openai_client():
    return openai.AsyncOpenAI(api_key="sk-fake123")


@pytest.fixture
def set_openai_response():
    def set_response(
        openai_mock: openai_responses.OpenAIMock, response: BaseModel, next_state: str
    ):
        openai_mock.beta.chat.completions.create.response = {
            "choices": [
                {
                    "index": 0,
                    "finish_reason": "stop",
                    "message": {
                        "content": wrap_into_json_response(response, next_state),
                        "role": "assistant",
                    },
                }
            ]
        }

    return set_response


# Test the module creation
def test_module_creation(fsm):
    """Test that the module can be created."""
    assert isinstance(fsm, MooreFSM)


# Test for simple fsm with state
@pytest.mark.asyncio
@openai_responses.mock()
async def test_config_data(
    fsm, openai_client, set_openai_response, openai_mock: openai_responses.OpenAIMock
):
    """Test that a contexts work"""

    COMMON_RESPONSE = "My name is Moore."
    CONTEXT_SET_DATA = {
        "name": "Moore",
        "age": 25,
    }

    @fsm.state(state_key="START", system_prompt="Hello, what's your name?")
    async def start_state(fsm: MooreFSM, response: str, will_transition: bool):
        assert response == COMMON_RESPONSE
        assert will_transition == False
        assert fsm.get_next_state() == "START"

        # Set the context data
        fsm.set_context_data("config", CONTEXT_SET_DATA)

    set_openai_response(
        openai_mock, DefaultResponse(content=COMMON_RESPONSE), next_state="START"
    )

    # Run the FSM
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "START"
    assert run_state.response == COMMON_RESPONSE

    # Check the context data
    assert run_state.context_data.get("config") == CONTEXT_SET_DATA
